from aiohttp import web
import json

__ws_list = None   # List of open websockets.
__subscriptions = {}
__log = None


def setup_websockets(app) -> list:
    global __ws_list
    global __log
    __log = app['log']
    if not isinstance(__ws_list, list):
        # very first run. We set up routes an install shutdown
        __ws_list = []
        app.router.add_get(app['options'].websocket_path,  _ws_handler)
        app.on_shutdown.append(_my_shutdown)
    return __ws_list


async def send_to_all(msg: dict) -> None:
    global __ws_list
    for ws in __ws_list:
        await ws.send_str(msg)
    return None


def subscribe(model: str, callback) -> None:
    """Install a callback function for given model.

    :param model: name of data model that asks for websocket
    :type model: str
    :param callback: callback function that is called when data arrives
    :type callback: function with parameter msg
    :return: None
    :rtype: None
    """
    global __subscriptions
    __subscriptions[model] = callback
    return None


async def _ws_handler(request) -> web.WebSocketResponse:
    global __subscriptions
    global __log
    global __ws_list
    ws = web.WebSocketResponse()
    await ws.prepare(request)

    __ws_list.append(ws)

# Here possiblity to send init-message to Web-Client

    try:
        async for msg in ws:
            __log.debug(msg)
            if msg.type == web.WSMsgType.TEXT:
                msg = json.loads(msg.data)
                await dispatch_message(msg)
            else:
                break
    except Exception as e:
        __log.error(f'websocket.py, exeption {e}')
    finally:
        __log.debug(f'Finally ws remove: {ws}')
        __ws_list.remove(ws)
    return ws


async def dispatch_message(msg: dict) -> bool:
    global __subscriptions
    global __log
    model = msg.get('model', None)
    if model is None:
        __log.error(f'Message not for a data model: {msg}')
        return False
    callback_func = __subscriptions.get(model, None)
    if callback_func is None:
        __log.error(f'Message for an unknown data model: {msg}')
        return False
    try:
        await callback_func(msg)
    except Exception as e:
        __log.error(f'websocket.py, exeption {e} in callable {callback_func}')
        return False
    return True


async def _my_shutdown(app) -> None:
    global __ws_list
    for ws in __ws_list:
        __log.debug(f'calling ws.close for: {ws}')
        await ws.close()
    return None



import asyncio
import configargparse
import pkg_resources
import os
import pathlib
# import logging
# import sys
import socket

import riego.database
import riego.valves
import riego.parameter
import riego.mqtt_gmqtt as riego_mqtt
import riego.logger
import riego.timer
import riego.boxes

from riego.web.routes import setup_routes
from riego.web.error_pages import setup_error_pages
from riego.web.websockets import setup_websockets

from aiohttp import web
import jinja2
import aiohttp_jinja2
import aiohttp_debugtoolbar

from riego.__init__ import __version__


async def on_startup(app):
    if app['options'].enable_asyncio_debug:
        asyncio.get_event_loop().set_debug(True)
    app['log'].info("on_startup")
# TODO bring next two lines to correspondending classes
    app['background_mqtt'] = asyncio.create_task(app['mqtt'].start_async())
    app['background_timer'] = asyncio.create_task(app['timer'].start_async())


async def on_shutdown(app):
    app['log'].info("on_shutdown")


async def on_cleanup(app):
    app['log'].info("on_cleanup")

    app['background_timer'].cancel()
    await app['background_timer']

    app['background_mqtt'].cancel()
    await app['background_mqtt']


def main():
    p = configargparse.ArgParser(
        default_config_files=['/etc/riego/conf.d/*.conf', '~/.riego.conf',
                              'riego.conf'])
    p.add('-c', '--config', is_config_file=True, env_var='RIEGO_CONF',
          required=False, help='config file path')
    p.add('-d', '--database', help='Path and name for DB file',
          default='db/riego.db')
    p.add('-e', '--event_log', help='Full path and name for event logfile',
          default='log/event.log')
    p.add('--event_log_max_bytes', help='Maximum Evet Log Size in bytes',
          default=1024*300, type=int)
    p.add('--event_log_backup_count', help='How many files to rotate',
          default=3, type=int)
    p.add('-m', '--mqtt_host', help='IP adress of mqtt host',
          default='127.0.0.1')
    p.add('-p', '--mqtt_port', help='Port of mqtt service',
          default=1883, type=int)
    p.add('--mqtt_client_id', help='Client ID for MQTT-Connection',
          default=f'riego_ctrl_{socket.gethostname()}')
    p.add('--mqtt_subscription_topic', help='MQTT Topic that we are listening',
          default='riego/#')
    p.add('--database_migrations_dir',
          help='path to database migrations directory',
          default=pkg_resources.resource_filename('riego', 'migrations'))
    p.add('--base_dir', help='Change only if you know what you are doing',
          default=pathlib.Path(__file__).parent)
    p.add('--http_server_bind_address',
          help='http-server bind address', default='0.0.0.0')
    p.add('--http_server_bind_port', help='http-server bind port',
          default=8080, type=int)
    p.add('--http_server_static_dir',
          help='Serve static html files from this directory',
          default=pkg_resources.resource_filename('riego.web', 'static'))
    p.add('--http_server_template_dir',
          help='Serve template files from this directory',
          default=pkg_resources.resource_filename('riego.web', 'templates'))
    p.add('--websocket_path', help='url path for websocket',
          default="/ws")
    p.add('--time_format', help='Store and display time',
          default="%Y-%m-%d %H:%M:%S")
    p.add('--mqtt_cmnd_prefix', help='',
          default="cmnd")
    p.add('--mqtt_result_subscription', help='',
          default="stat/+/RESULT")
    p.add('--mqtt_status_subscription', help='yet not used',
          default="stat/+/STATUS")
    p.add('--mqtt_tele_subscription', help='yet not used',
          default="tele/#")
    p.add('--mqtt_tele_filter_LWT', help='yet not used',
          default="tele/+/LWT")
    p.add('--mqtt_tele_filter_SENSOR',
          help='yet not used', default="tele/+/SENSOR")
    p.add('--mqtt_tele_filter_state', help='yet not used',
          default="tele/+/state")
    p.add('--mqtt_keyword_ON', help='',
          default="ON")
    p.add('--mqtt_keyword_OFF', help='',
          default="OFF")

    p.add('--enable_aiohttp_debug_toolbar', action='store_true')
    p.add('--enable_asyncio_debug', action='store_true')
    p.add('--enable_timer_dev_mode', action='store_true')

    p.add('-v', '--verbose', help='verbose', action='store_true')
    p.add('--version', help='Print version and exit', action='store_true')
    p.add('--defaults', help='Print options with default values and exit',
          action='store_true')

    options = p.parse_args()

    try:
        with open('riego.conf', 'xt') as f:
            for item in vars(options):
                f.write(f'# {item}={getattr(options, item)}\n')
    except IOError:
        pass

    if options.defaults:
        for item in vars(options):
            print(f'# {item}={getattr(options, item)}')
        exit(0)

    if options.version:
        print('Version: ', __version__)
        exit(0)

    if options.verbose:
        print(p.format_values())

#    if sys.version_info >= (3, 8):
#        asyncio.DefaultEventLoopPolicy = asyncio.WindowsSelectorEventLoopPolicy  # noqa: E501

    if os.name == "posix":
        import uvloop  # pylint: disable=import-error
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

    app = web.Application()

    app['version'] = __version__
    app['options'] = options
    app['log'] = riego.logger.create_log(options)
    app['event_log'] = riego.logger.create_event_log(options)
    app['db'] = riego.database.Database(app)
    app['mqtt'] = riego_mqtt.Mqtt(app)
    app['valves'] = riego.valves.Valves(app)
    app['parameter'] = riego.parameter.Parameter(app)
    app['timer'] = riego.timer.Timer(app)
    app['boxes'] = riego.boxes.Boxes(app)

    app.on_startup.append(on_startup)
    app.on_shutdown.append(on_shutdown)
    app.on_cleanup.append(on_cleanup)

    aiohttp_jinja2.setup(app,
                         loader=jinja2.FileSystemLoader(
                             options.http_server_template_dir))

    setup_routes(app)
    setup_error_pages(app)
    setup_websockets(app)

    if options.enable_aiohttp_debug_toolbar:
        aiohttp_debugtoolbar.setup(app, check_host=False)

    app['log'].info("Start")
    app['event_log'].info('Start')

    web.run_app(app,
                host=options.http_server_bind_address,
                port=options.http_server_bind_port)



import os
import logging
import time

import asyncio
import mqttools


class Mqtt:
    def __init__(self, app):
        self.__options = app['options']
        self.__log = app['log']

        self.client = mqttools.Client(self.__options.mqtt_host,
                                      self.__options.mqtt_port,
                                      client_id=self.__options.mqtt_client_id)

    async def start_async(self):

        await self.client.start()
        await self.client.subscribe(self.__options.mqtt_subscription_topic)

        try:
            while True:
                topic, message = await self.client.messages.get()

                if topic is None:
                    print('Broker connection lost!')
                    break

                print(f'Topic:   {topic}')
                print(f'Message: {message}')
        except asyncio.CancelledError:
            await self.client.disconnect()

#        while True:
#            try:
#                await asyncio.sleep(1)
#            except asyncio.CancelledError:
#                self.__log.debug('MQTT: trapped cancel')
#                break
#        self.__log.debug('MQTT: call disconnect()')
#        await self.client.disconnect()


class Options():

    def __init__(self):
        self.mqtt_host = "192.168.88.229"
        self.mqtt_port = 1883
        self.mqtt_client_id = "testing"
        self.mqtt_subscription_topic = "riego/#"


async def main():
    app = {}
    options = Options()
    app['options'] = options
    app['log'] = logging.getLogger()

    mqttc = Mqtt(app)
    await mqttc.start_async()

    mqttc.client.publish('riego/TEST/TIME', str(time.time()), qos=1)

    await asyncio.sleep(6)
    await mqttc.client.disconnect()


if __name__ == "__main__":
    if os.name == 'nt':
        asyncio.DefaultEventLoopPolicy = asyncio.WindowsSelectorEventLoopPolicy
    asyncio.run(main())



import asyncio
import logging
import re

from gmqtt import Client as MQTTClient
from gmqtt.mqtt.constants import MQTTv311


class Mqtt:
    def __init__(self, app):
        self._options = app['options']
        self._log = app['log']
        self.client = MQTTClient(self._options.mqtt_client_id)
        self.client.on_connect = self._on_connect
        self.client.on_message = self._on_message
        self.client.on_disconnect = self._on_disconnect
        self.client.on_subscribe = self._on_subscribe
        self.subscriptions = {}

    def _on_connect(self, client, flags, rc, properties):
        self._log.debug('MQTT Connected')

        for key in self.subscriptions:
            self.client.subscribe(key, qos=0)

    async def _on_message(self, client, topic, payload, qos, properties):
        payload = payload.decode()
#        self._log.debug(f'MQTT RECV MSG: {payload}, TOPIC: {topic}')
        for key in self.subscriptions:
            if self.match_topic(topic, key):
                func = self.subscriptions[key]
                await func(topic, payload)
        return 0

    def _on_disconnect(self, client, packet, exc=None):
        self._log.debug('MQTT Disconnected')

    def _on_subscribe(self, client, mid, qos, properties):
        self._log.debug('MQTT SUBSCRIBED')

    async def start_async(self):
        await self.client.connect(self._options.mqtt_host,
                                  port=self._options.mqtt_port,
                                  keepalive=10,
                                  version=MQTTv311)
#        while True:
#            try:
#                await asyncio.sleep(1)
#            except asyncio.CancelledError:
#                self.__log.debug('MQTT: trapped cancel')
#                break
#        self.__log.debug('MQTT: call disconnect()')
#        await self.client.disconnect()

    def subscribe(self, topic: str, callback: callable) -> None:
        self.subscriptions[topic] = callback

#    Der MQTT Client ist noch nicht gestartet.
#    Subscribe ist noch nicht möglich.
#        self.client.subscribe(topic, qos=0)
        return None

    def match_topic(self, topic: str, sub: str) -> bool:
        # Achtung auch "stat/+/RESULT" und "stat/+/RESULT1234"
        # ergebn ein Match! Die regex ist nicht korrekt!
        ret = bool(re.match(sub.translate({43: "[^/]+", 35: ".+"}), topic))
        return ret

    async def shutdown(self):
        await self.client.disconnect()


# Manually create Mock-Object


class Options():

    def __init__(self):
        self.mqtt_host = "192.168.88.229"
        self.mqtt_port = 1883
        self.mqtt_client_id = "testing"


async def main():
    app = {}
    options = Options()
    app['options'] = options
    logger = logging
    logging.basicConfig(level=logging.DEBUG)
    app['log'] = logger

    mqttc = Mqtt(app)
    await mqttc.start_async()

    mqttc.client.publish('cmnd/dev01/POWER1', payload="ON", qos=1)

    try:
        await asyncio.sleep(60)
    except asyncio.CancelledError:
        print('MQTT: trapped cancel')
    finally:
        await mqttc.client.disconnect()


if __name__ == "__main__":
    asyncio.run(main())





class Boxes():
    def __init__(self, app):
        self.mqtt = app['mqtt']
        self.mqtt.subscribe('tele/#', self._mqtt_handler)

    async def _mqtt_handler(self, topic: str, payload: str) -> bool:
        print(f'Topic: {topic}, payload: {payload}')
        return True



import socket
import paho.mqtt.client as mqtt
import asyncio


import time
import logging
import os


class AsyncioHelper:
    def __init__(self, loop, client):
        self.loop = loop
        self.client = client
        self.client.on_socket_open = self.on_socket_open
        self.client.on_socket_close = self.on_socket_close
        self.client.on_socket_register_write = self.on_socket_register_write
        self.client.on_socket_unregister_write = self.on_socket_unregister_write

    def on_socket_open(self, client, userdata, sock):
        print("Socket opened")

        def cb():
            print("Socket is readable, calling loop_read")
            client.loop_read()

        self.loop.add_reader(sock, cb)
        self.misc = self.loop.create_task(self.misc_loop())

    def on_socket_close(self, client, userdata, sock):
        print("Socket closed")
        self.loop.remove_reader(sock)
        self.misc.cancel()

    def on_socket_register_write(self, client, userdata, sock):
        print("Watching socket for writability.")

        def cb():
            print("Socket is writable, calling loop_write")
            client.loop_write()

        self.loop.add_writer(sock, cb)

    def on_socket_unregister_write(self, client, userdata, sock):
        print("Stop watching socket for writability.")
        self.loop.remove_writer(sock)

    async def misc_loop(self):
        print("misc_loop started")
        while self.client.loop_misc() == mqtt.MQTT_ERR_SUCCESS:
            try:
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                break
        print("misc_loop finished")


class Mqtt:
    def __init__(self, app):
        self.__options = app['options']
        self.__log = app['log']
        self.__loop = asyncio.get_event_loop()
        self.got_message = None

        self.client = mqtt.Client(client_id=self.__options.mqtt_client_id)
        self.client.on_connect = self._on_connect
        self.client.on_message = self._on_message
        self.client.on_disconnect = self._on_disconnect

        aioh = AsyncioHelper(self.__loop, self.client)

    def _on_connect(self, client, userdata, flags, rc):
        self.__log.debug('MQTT Connected')
        client.subscribe('TEST/#', qos=0)

    def _on_message(self, client, userdata, msg):
        self.__log.debug('MQTT RECV MSG:' + str(msg))
        print('MQTT RECV MSG:' + str(msg))
#        if not self.got_message:
#            print("Got unexpected message: {}".format(msg.decode()))
#        else:
#            self.got_message.set_result(msg.payload)

    def _on_disconnect(self, client, userdata, rc):
        self.disconnected.set_result(rc)
        self.__log.debug('MQTT Disconnected')
#        exit(1)

    async def start_async(self):

        self.client.connect(self.__options.mqtt_host,
                            self.__options.mqtt_port, 10)
        self.client.socket().setsockopt(socket.SOL_SOCKET,
                                        socket.SO_SNDBUF,
                                        2048)

        while True:
            try:
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                self.__log.debug('MQTT: trapped cancel')
                break
        self.__log.debug('MQTT: call disconnect()')
        await self.client.disconnect()


class Options():

    def __init__(self):
        self.mqtt_host = "192.168.88.229"
        self.mqtt_port = 1883
        self.mqtt_client_id = "testing"


async def main():
    app = {}
    options = Options()
    app['options'] = options
    app['log'] = logging.getLogger()

    mqttc = Mqtt(app)
    await mqttc.start_async()

    mqttc.client.publish('TEST/TIME', str(time.time()), qos=1)

    await asyncio.sleep(6)
    await mqttc.client.disconnect()


if __name__ == "__main__":
    if os.name == 'nt':
        asyncio.DefaultEventLoopPolicy = asyncio.WindowsSelectorEventLoopPolicy
    asyncio.run(main())



"""

"""

from yoyo import step

__depends__ = {'__init__'}

steps = [
    step('''CREATE TABLE "valves" (
    "id"	INTEGER,
    "name"	TEXT UNIQUE,
    "remark"	TEXT,
    "box_id"	INTEGER NOT NULL,
    "channel"	TEXT NOT NULL,
    "duration"	INTEGER DEFAULT 0,
    "interval"	INTEGER DEFAULT 0,
    "last_run"	INTEGER DEFAULT '1970-01-01 00:000:00',
    "is_running"	INTEGER DEFAULT 0,
    "is_enabled"	INTEGER DEFAULT 0,
    UNIQUE("channel","box_id"),
    PRIMARY KEY("id"))''',
         '''DROP TABLE "valves" '''
         ),

    step('''CREATE TABLE "boxes" (
    "id"	INTEGER,
    "topic"	TEXT UNIQUE,
    "hostname"	TEXT,
    "display_name"	TEXT UNIQUE,
    "remark"	TEXT,
    "first_seen"	TEXT,
    "has_config"	INTEGER,
    "online_since"	TEXT,
    "hw_type"	TEXT,
    "hw_version"	TEXT,
    "sw_version"	TEXT,
    "fallback_topic"	TEXT,
    "group_topic"	TEXT,
    "ip_address"	TEXT,
    PRIMARY KEY("id"))''',
         '''DROP TABLE "boxes" '''
         ),

    step('''CREATE TABLE "parameter" (
    "id"	INTEGER,
    "key"	TEXT UNIQUE,
    "value"	TEXT,
    PRIMARY KEY("id"))''',
         '''DROP TABLE "parameter" '''
         ),

    step('''CREATE TABLE "event_log" (
    "id"	INTEGER,
    "event"	TEXT,
    "date"	TEXT,
    PRIMARY KEY("id"))''',
         '''DROP TABLE "event_log" '''
         ),

    step(
        "INSERT INTO parameter (key,value) VALUES ('startTime', '19:00')",
        "DELETE FROM parameter WHERE key = 'startTime'"
    ),
    step(
        "INSERT INTO parameter (key,value) VALUES ('maxDuration', '240')",
        "DELETE FROM parameter WHERE key = 'maxDuration'"
    ),
    step(
        "INSERT INTO parameter (key,value) VALUES ('heat', '0')",
        "DELETE FROM parameter WHERE key = 'heat'"
    ),
    step(
        "INSERT INTO parameter (key,value) VALUES ('rain', '0')",
        "DELETE FROM parameter WHERE key = 'rain'"
    ),
    step(
        "INSERT INTO parameter (key,value) VALUES ('pause', '0')",
        "DELETE FROM parameter WHERE key = 'pause'"
    ),
    step(
        "INSERT INTO parameter (key,value) VALUES ('disable', '0')",
        "DELETE FROM parameter WHERE key = 'enable'"
    )


]



import aiohttp_jinja2
from aiohttp import web


import asyncio
import sys
import json

from riego.__init__ import __version__
from pkg_resources import packaging


@aiohttp_jinja2.template('system/index.html')
async def system_index(request):
    installed_version = await _check_installed()
    if not packaging.version.parse(installed_version) == packaging.version.parse(__version__):  # noqa: E501
        text = '''Diese Riego Instanz läuft in der Version {}
                und entspricht nicht der installierten Version {}.'''  # noqa: E501
        text = text.format(__version__, installed_version)

    return {'text': text}


@aiohttp_jinja2.template('system/index.html')
async def system_check_update(request):
    update = await _check_update("No update")
    return {'text':  update}


@aiohttp_jinja2.template('system/index.html')
async def system_do_update(request):
    await _do_update()
    return {'text': "Restart erforderlich"}


@aiohttp_jinja2.template('system/index.html')
async def system_restart(request):
    exit(0)


@aiohttp_jinja2.template('system/index.html')
async def system_setup(request):
    return {'text': 'Hello, world2'}


@aiohttp_jinja2.template('system/index.html')
async def system_event_log(request):
    with open(request.app['options'].event_log, "rt") as fp:
        # return web.Response(body=fp.read(), content_type="text/plain")
        return {'text': fp.read()}


async def system_exc(request):
    raise NotImplementedError


async def _check_installed():
    version = "0.0.0"
    proc = await asyncio.create_subprocess_exec(
        sys.executable, '-m', 'pip', 'list', "--format=json",
        "--disable-pip-version-check",
        "--no-color",
        "--no-python-version-warning",
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE)

    stdout, stderr = await proc.communicate()
    data = json.loads(stdout)
    for item in data:
        if item['name'] == 'riego':
            version = item['version']
            break
    return version


async def _check_update(latest_version=None):
    proc = await asyncio.create_subprocess_exec(
        sys.executable, '-m', 'pip', 'list', "-o", "--format=json",
        "--disable-pip-version-check",
        "--no-color",
        "--no-python-version-warning",
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE)

    stdout, stderr = await proc.communicate()
    data = json.loads(stdout)
    for item in data:
        if item['name'] == 'riego':
            latest_version = item['latest_version']
            break
    return latest_version


async def _do_update():
    proc = await asyncio.create_subprocess_exec(
        sys.executable, '-m', 'pip', 'install', "riego", "--upgrade",
        "--disable-pip-version-check",
        "--no-color",
        "--no-python-version-warning",
        "-q", "-q", "-q")
    await proc.wait()
    return proc.returncode


async def _stream_rsponse(request: web.Request) -> web.StreamResponse:
    resp = web.StreamResponse()
    name = request.match_info.get("name", "Anonymous")
    answer = ("Hello, " + name).encode("utf8")
    resp.content_length = len(answer)
    resp.content_type = "text/plain"
    await resp.prepare(request)
    await resp.write(answer)
    await resp.write_eof()
    return resp



import paho.mqtt.client
import asyncio
import socket

_instance = None


class Mqtt:
    def __init__(self, app):
        global _instance
        _instance = self
        options = app['options']
        self.__log = app['log']
        self.client = paho.mqtt.client.Client()
        self.client.on_connect = self._on_connect
        self.client.on_disconnect = self._on_disconnect
        self.client.on_message = self._on_message
        self.client.connect(options.mqtt_host, options.mqtt_port, 10)
        self.subscriptions = {}
        self.__wait_for_first_conn()

    def __wait_for_first_conn(self):
        self.client.loop()
        self.client.loop()

    def _on_connect(self, client, userdata, flags, rc):
        if rc == 0:
            self.__log.info("Mqtt Connection established")
            for key in self.subscriptions:
                self.client.subscribe(key)
                self.__log.debug("Mqtt subribed to: " + str(key))

        else:
            self.__log.critical("Bad mqtt connection, rc: " + str(rc))

    # The callback for when a PUBLISH message is received from the server.

    def _on_message(self, client, userdata, msg):
        for key in self.subscriptions:
            if key == msg.topic:
                func = self.subscriptions[key]
#                func(msg)
                self.__log.debug("callback" + msg.topic+" "+str(msg.payload))

    def _on_disconnect(self, client, userdata, rc):
        if rc != 0:
            self.__log.error(
                "Unexpected mqtt disconnection rc: " + str(rc))

    def subscribe(self, topic: str, callback) -> None:
        self.subscriptions[topic] = callback
        for key in self.subscriptions:
            print(key)
        self.client.subscribe(topic)
        return None

    async def start_async(self):
        self.client.socket().setsockopt(socket.SOL_SOCKET,
                                        socket.SO_SNDBUF,
                                        2048)
        while True:
            try:
                # TODO: convert to Task
                self.client.loop()
                await asyncio.sleep(1)
            except asyncio.CancelledError:
                self.__log.debug("mqtt: trapped cancel")
                break
        self.client.disconnect()


def get_mqtt() -> Mqtt:
    global _instance
    return _instance



import sqlite3
from pathlib import Path
from yoyo import read_migrations
from yoyo import get_backend


class Database:
    def __init__(self, app):
        options = app['options']
        self.log = app['log']
        self.conn = None
        Path(options.database).parent.mkdir(parents=True, exist_ok=True)
        try:
            backend = get_backend('sqlite:///' + options.database)
        except Exception as e:
            self.log.error(f'Not able to open database: {e}')
            exit(1)

        migrations = read_migrations(options.database_migrations_dir)

        with backend.lock():
            backend.apply_migrations(backend.to_apply(migrations))
        try:
            self.conn = sqlite3.connect(options.database)
            self.conn.row_factory = sqlite3.Row
        except Exception as e:
            self.log.error(f'Not able to connect to database: {e}')
            if self.conn is not None:
                self.conn.close()
            exit(1)

    def __del__(self):
        try:
            if self.conn is not None:
                self.conn.close()
        except Exception as e:
            self.log.error(f'database.py: not able to close conn: {e}')



from datetime import datetime, timedelta
import asyncio


class Timer():
    def __init__(self, app):
        self._valves = app['valves']
        self._parameter = app['parameter']
        self._options = app['options']
        self._log = app['log']

        self._maxDuration = int(self._parameter.get('maxDuration'))
        self._start_hour, self._start_minute = self._parameter.get(
            'startTime').split(':')
        self._start_hour = int(self._start_hour)
        self._start_minute = int(self._start_minute)
        self._running_period_end = None
        self._running_period_start = None

    async def start_async(self):
        v = self._valves.get_next()
        while True:
            try:
                if v is None:
                    # No valves in database found, system is initalizing
                    await asyncio.sleep(2)
                    v = self._valves.get_next()
                    continue
                if v.is_running == -1:
                    v = self._valves.get_next()
                    continue
                if v.is_running == 1:
                    if self._options.enable_timer_dev_mode:
                        td = timedelta(minutes=0, seconds=v.duration)
                    else:
                        td = timedelta(minutes=v.duration)

                    if datetime.now() - v.last_run > td:
                        # Laufzeit erreicht
                        await v.set_is_running(0)
                        self._log.debug("valveOff " + v.name)
                    else:
                        pass
                if v.is_running == 0:
                    if self._options.enable_timer_dev_mode:
                        td = timedelta(days=0, seconds=v.interval)
                    else:
                        td = timedelta(days=v.interval)

                    if (datetime.now() - v.last_run > td and v.is_enabled and
                            await self.is_running_period()):
                        # Intervall erreicht
                        await v.set_is_running(1)
                        self._log.debug("valveOn " + v.name)
                    else:
                        v = self._valves.get_next()
                        self._log.debug("NextValve: " + v.name)
                await asyncio.sleep(2)
            except asyncio.CancelledError:
                self._log.debug("Timer: trapped cancel")
                break
        self._log.debug("Timer: shutdown valve")
        if v is not None:
            await v.set_is_running(0)

    async def is_running_period(self) -> bool:
        print("in function is_running_period")
        if self._running_period_start is None:
            self._running_period_start = datetime.now().replace(
                hour=self._start_hour, minute=self._start_minute)

        if datetime.now() < self._running_period_start:
            self._running_period_end = None
            return False

        if self._running_period_end is None:
            self._running_period_end = self._running_period_start + \
                timedelta(minutes=self._maxDuration)

        if datetime.now() > self._running_period_end:
            self._running_period_start = None
            return False
        print("is running period")
        return True



from datetime import datetime
import json
import riego.web.websockets
import re

bool_to_int = {'true': 1, 'false': 0, True: 1, False: 0,
               'True': 1, 'False': 0, 'on': 1, 'off': 0,
               'On': 1, 'Off': 0, 'ON': 1, 'OFF': 0}

int_to_js_bool = {1: "true", 0: "false", -1: "-1"}


class Valve():
    def __init__(self, row, app):
        self.__db_conn = app['db'].conn
        self.__mqtt = app['mqtt']
        self.__event_log = app['event_log']
        self.__options = app['options']

        self.__id = row['id']
        self.__name = row['name']
        self.__remark = row['remark']

        self.__channel = row['channel']
        self.__topic = row['box_topic'] + "/" + row['channel']
        self.__duration = row['duration']
        self.__interval = row['interval']
        self.__last_run = row['last_run']
        self.__is_running = row['is_running']
        self.__is_enabled = row['is_enabled']

        self.__box_display_name = row['box_display_name']
        self.__box_id = row['box_id']

    @ property
    def id(self):
        return self.__id

    @ property
    def name(self):
        return self.__name

    async def set_name(self, val: str):
        with self.__db_conn:
            self.__db_conn.execute(
                'UPDATE valves SET name = ?  WHERE id = ?',
                (val, self.__id))
        await self.send_status_with_websocket('name', val)
        self.__name = val

    @ property
    def topic(self):
        return self.__topic

    async def set_topic(self, val) -> NotImplementedError:
        raise NotImplementedError

    @ property
    def duration(self):
        return self.__duration

    async def set_duration(self, val: int):
        val = int(val)
        with self.__db_conn:
            self.__db_conn.execute(
                'UPDATE valves SET duration = ?  WHERE id = ?',
                (val, self.__id))
        await self.send_status_with_websocket('duration', val)
        self.__duration = val

    @ property
    def interval(self):
        return self.__interval

    async def set_interval(self, val: int):
        val = int(val)
        with self.__db_conn:
            self.__db_conn.execute(
                'UPDATE valves SET interval = ?  WHERE id = ?',
                (val, self.__id))
        await self.send_status_with_websocket('interval', val)
        self.__interval = val

    @ property
    def last_run(self) -> datetime:
        """Convert to datetime object before return

        :return: Last ON-Time of valve
        :rtype: datetime
        """
        return datetime.strptime(self.__last_run, self.__options.time_format)

    async def set_last_run(self):
        raise NotImplementedError

    @ property
    def is_running(self):
        return self.__is_running

    async def set_is_running(self, val):
        val = bool_to_int[val]
        if val:
            await self._set_on_try()
        else:
            await self._set_off_try()

    @ property
    def is_enabled(self):
        return self.__is_enabled

    async def set_is_enabled(self, val):
        val = bool_to_int[val]

        with self.__db_conn:
            self.__db_conn.execute(
                'UPDATE valves SET is_enabled = ?  WHERE id = ?',
                (val, self.__id))
        await self.send_status_with_websocket('is_enabled', val)
        self.__is_enabled = val

    @ property
    def remark(self):
        return self.__remark

    async def set_remark(self, val):
        with self.__db_conn:
            self.__db_conn.execute(
                'UPDATE valves SET remark = ?  WHERE id = ?',
                (val, self.__id))
        await self.send_status_with_websocket('remark', val)
        self.__remark = val

    def get_dict(self):
        ret = {'id': self.__id,
               'name': self.__name,
               'topic': self.__topic,
               'duration': self.__duration,
               'interval': self.__interval,
               'last_run': self.__last_run,
               'is_running': self.__is_running,
               'is_enabled': self.__is_enabled,
               'remark': self.__remark}
        return ret

    async def send_status_with_websocket(self, prop, value):
        ret = {
            'action': "status",
            'model': "valves",
            'id': self.__id,
            'prop': prop,
            'value': value,
        }
        ret = json.dumps(ret)
        await riego.web.websockets.send_to_all(ret)
        return ret

    async def _set_on_try(self) -> bool:
        self.__is_running = -1
        with self.__db_conn:
            self.__db_conn.execute(
                'UPDATE valves SET is_running = ?  WHERE id = ?',
                (self.__is_running, self.__id))
        topic = self.__options.mqtt_cmnd_prefix + '/' + self.__topic
        await self.send_status_with_websocket('is_running', -1)
        if self.__mqtt.client is None:
            return False
        if not self.__mqtt.client.is_connected:
            return False
        self.__mqtt.client.publish(topic, 1)
        return True

    async def _set_off_try(self) -> bool:
        self.__is_running = -1
        with self.__db_conn:
            self.__db_conn.execute(
                'UPDATE valves SET is_running = ?  WHERE id = ?',
                (self.__is_running, self.__id))
        await self.send_status_with_websocket('is_running', -1)
        if self.__mqtt.client is None:
            return False
        if not self.__mqtt.client.is_connected:
            return False
        topic = self.__options.mqtt_cmnd_prefix + '/' + self.__topic
        self.__mqtt.client.publish(topic, 0)
        return True

    async def _set_on_confirm(self):
        self.__is_running = 1
        with self.__db_conn:
            self.__db_conn.execute(
                'UPDATE valves SET is_running = ?  WHERE id = ?',
                (self.__is_running, self.__id))
        await self.send_status_with_websocket('is_running', 1)

        self.__last_run = datetime.now().strftime(self.__options.time_format)
        with self.__db_conn:
            self.__db_conn.execute(
                'UPDATE valves SET last_run = ?  WHERE id = ?',
                (self.__last_run, self.__id))
        await self.send_status_with_websocket('last_run', self.__last_run)

    async def _set_off_confirm(self):
        self.__is_running = 0
        with self.__db_conn:
            self.__db_conn.execute(
                'UPDATE valves SET is_running = ?  WHERE id = ?',
                (self.__is_running, self.__id))
        await self.send_status_with_websocket('is_running', 0)


class Valves():
    def __init__(self, app):
        db_conn = app['db'].conn
        mqtt = app['mqtt']
        self.log = app['log']
        self.options = app['options']

        self._valves = []
        self.idx_valves = -1
        sql = '''SELECT valves.*, 
            boxes.id AS box_id,
            boxes.topic AS box_topic,
            boxes.display_name AS box_display_name
            FROM valves, boxes 
            WHERE valves.box_id = boxes.id;'''
        for row in db_conn.execute(sql):
            v = Valve(row, app)
            self._valves.append(v)
# TODO Dependenca Injection for websockets.
        riego.web.websockets.subscribe('valves', self._ws_handler)
        mqtt.subscribe(self.options.mqtt_result_subscription,
                       self._mqtt_result_handler)

    def get_next(self):
        self.idx_valves += 1
        if len(self._valves) == 0:
            return None
        if self.idx_valves >= len(self._valves):
            self.idx_valves = 0
        return self._valves[self.idx_valves]

    def get_dict_of_all(self) -> dict:
        """Create dict of properties and value from all
        objects of valves. used for updating HTML-page

        :return: dict with all properties of all valves
        :rtype: dict
        """
        ret = {}
        for v in self._valves:
            ret[v.id] = v.get_dict()
        return ret

    def get_valve_by_id(self, id: int) -> Valve:
        """Returns Valve-Object from database with given id.

        :param id: Unique id from database
        :type id: int
        :return: None if not found
        :rtype: Valve
        """
        for v in self._valves:
            if v.id == int(id):
                return v
        return None

    def get_valve_by_topic(self, topic: str) -> Valve:
        """Returns Valve-Object from database with given topic

        :param topic: Unique topic from database
        :type topic: str
        :return: None if not found
        :rtype: Valve
        """
        for v in self._valves:
            if v.topic == topic:
                return v
        return None

    async def _ws_handler(self, msg: dict):
        """Find object from class valve with id=msg[id] and
        Call setter-method of Class Valve according to msg['prop']
        """
        self.log.debug(f'In Valves._ws_handler: {msg}')

        if not msg['action'] == 'update':
            return None
        valve = self.get_valve_by_id(msg['id'])
        func = getattr(valve, "set_" + msg['prop'])
        await func(msg['value'])

    async def _mqtt_result_handler(self, topic: str, payload: str) -> bool:
        """Dispatch mqtt message "stat/box_name/RESULT {POWER1 :  ON}"


        :param topic: Topic of subscribed MQTT-message
        :type topic: str
        :param payload: Payload of subscribed MQTT-message
        :type payload: str
        :return: [description]
        :rtype: bool
        """
        box = re.search('/(.*?)/', topic).group(1)
        payload = json.loads(payload)
        for channel in payload:
            topic = box + '/' + channel
            valve = self.get_valve_by_topic(topic)
            if valve is None:
                self.log.error(f'valves._mqtt_result_handler: unknown topic: {topic}')  # noqa: E501
                continue
            value = bool_to_int[payload[channel]]
            if value == 1:
                await valve._set_on_confirm()
            else:
                await valve._set_off_confirm()
        return True



from riego.web.dashboard.views import dashboard_index
from riego.web.system.views import system_index, system_check_update, system_do_update, system_restart, system_setup, system_event_log, system_exc


routes = [
    ('GET', '/',                    dashboard_index,      'dashboard_index'),
    ('GET', '/system',              system_index,         'system_index'),
    ('GET', '/system/check_update', system_check_update,  'system_check_update'),
    ('GET', '/system/do_update',    system_do_update,     'system_do_update'),
    ('GET', '/system/restart',      system_restart,       'system_restart'),
    ('GET', '/system/setup',        system_setup,         'system_setup'),
    ('GET', '/system/event_log',    system_event_log,     'system_event_log'),
    ('GET', '/exc',                 system_exc,           'system_exc'),
]


def setup_routes(app):
    global routes

    for route in routes:
        app.router.add_route(route[0], route[1], route[2], name=route[3])

    app.router.add_static('/static', app['options'].http_server_static_dir,
                          name='static', show_index=True)
